from typing import Union

from ..._types import TypedDict


class Seller(TypedDict):
    fulfillment_seller_id: Union[str, None]
    is_active: bool
    is_better_scope: bool
    merchant_name: Union[str, None]
    name: str
    seller_id: str
    seller_type: int


class SalesChannel(TypedDict):
    id: int
    is_active: bool
    name: str
    position: Union[int, None]
    time_zone: str



from http import HTTPStatus
from json import JSONDecodeError
from time import sleep
from typing import Any, Type, Union, cast

from httpx import (
    Client,
    Headers,
    HTTPError,
    HTTPStatusError,
    Response,
)
from httpx._types import (
    CookieTypes,
    HeaderTypes,
    QueryParamTypes,
    RequestContent,
    RequestData,
    RequestFiles,
)
from limits import RateLimitItem, storage, strategies
from tenacity import (
    retry,
    retry_if_exception_type,
    stop_after_attempt,
    wait_exponential,
)

from .._config import VTEXConfig
from .._constants import APP_KEY_HEADER, APP_TOKEN_HEADER
from .._dto import VTEXDataResponse, VTEXResponseType
from .._exceptions import VTEXRequestError, VTEXResponseError
from .._logging import disable_loggers, get_logger, log_before_retry
from .._types import HTTPMethodType
from .._utils import join_url, redact_headers, to_snake_case, to_snake_case_deep
from .._vtex import VTEX


class BaseAPI:
    """
    Base client for a VTEX API.
    """

    def __init__(self, client: VTEX) -> None:
        self.client = client
        self.logger = get_logger(
            name=to_snake_case(type(self).__name__),
            parent=self.client.logger,
        )

    def _request(
        self,
        method: HTTPMethodType,
        endpoint: str,
        environment: Union[str, None] = None,
        headers: Union[HeaderTypes, None] = None,
        cookies: Union[CookieTypes, None] = None,
        params: Union[QueryParamTypes, None] = None,
        json: Union[Any, None] = None,
        data: Union[RequestData, None] = None,
        content: Union[RequestContent, None] = None,
        files: Union[RequestFiles, None] = None,
        config: Union[VTEXConfig, None] = None,
        rate_limit: Union[RateLimitItem, None] = None,
        response_class: Union[Type[VTEXResponseType], None] = None,
        **kwargs: Any,
    ) -> VTEXResponseType:
        request_config = (config or self.client.config).with_overrides(**kwargs)

        if environment:
            url = join_url(
                f"https://{request_config.account_name}.{environment}.com.br",
                endpoint,
            )
        else:
            url = join_url(
                "https://api.vtex.com/",
                request_config.account_name,
                endpoint,
            )

        headers = Headers(headers=headers)
        headers[APP_KEY_HEADER] = request_config.app_key.get_secret_value()
        headers[APP_TOKEN_HEADER] = request_config.app_token.get_secret_value()
        headers["Content-Type"] = "application/json; charset=utf-8"
        headers["Accept"] = "application/json"

        rate_limiter = (
            None
            if not (rate_limit and request_config.rate_limit_storage_url)
            else strategies.MovingWindowRateLimiter(
                storage.storage_from_string(request_config.rate_limit_storage_url),
            )
        )

        @retry(
            stop=stop_after_attempt(
                max_attempt_number=request_config.retry_attempts + 1,
            ),
            wait=wait_exponential(
                min=request_config.retry_backoff_min,
                max=request_config.retry_backoff_max,
                exp_base=2.0,
            ),
            retry=retry_if_exception_type(exception_types=HTTPStatusError),
            before_sleep=(
                log_before_retry(
                    logger=self.logger,
                    log_level=request_config.log_retries,
                    environment=environment,
                    endpoint=endpoint,
                    account_name=request_config.account_name,
                )
                if request_config.log_retries is not False
                else None
            ),
            reraise=True,
        )
        def send_vtex_request() -> Response:
            with Client(timeout=request_config.timeout) as client:
                with disable_loggers(["httpcore", "httpx"]):
                    while (
                        rate_limiter
                        and rate_limit
                        and not rate_limiter.test(
                            rate_limit,
                            request_config.account_name,
                            endpoint,
                        )
                    ):
                        sleep(1)
                        self.logger.info(
                            f"{rate_limiter} limited {request_config.account_name} on "
                            f"{endpoint}",
                            extra={
                                "endpoint": endpoint,
                                "account_name": request_config.account_name,
                                "rate_limiter": rate_limiter,
                                "rate_limit": rate_limit,
                            },
                        )

                    try:
                        response = client.request(
                            method.upper(),
                            url,
                            headers=headers,
                            cookies=cookies,
                            params=params,
                            json=json,
                            data=data,
                            content=content,
                            files=files,
                        )
                    finally:
                        if rate_limiter and rate_limit:
                            rate_limiter.hit(
                                rate_limit,
                                request_config.account_name,
                                endpoint,
                            )

                response.request.headers = Headers(
                    redact_headers(dict(response.request.headers)),
                )
                response.headers = Headers(redact_headers(dict(response.headers)))

                for should_log, log_level in [
                    (response.is_informational, request_config.log_1xx),
                    (response.is_success, request_config.log_2xx),
                    (response.is_redirect, request_config.log_3xx),
                    (response.is_client_error, request_config.log_4xx),
                    (response.is_server_error, request_config.log_5xx),
                ]:
                    if should_log and log_level:
                        self.logger.log(
                            log_level,
                            f"{response.request.method} {response.request.url} "
                            f"{response.status_code} {response.reason_phrase}",
                            extra={
                                "method": response.request.method,
                                "url": response.request.url,
                                "status_code": response.status_code,
                                "environment": environment,
                                "endpoint": endpoint,
                                "account_name": request_config.account_name,
                            },
                        )

                if response.status_code in request_config.retry_status_codes:
                    response.raise_for_status()

                return response

        try:
            response = send_vtex_request()
        except HTTPStatusError as exception:
            response = exception.response
        except HTTPError as exception:
            headers = redact_headers(dict(headers))

            details = {
                "exception": exception,
                "method": str(method).upper(),
                "url": str(url),
                "headers": headers,
            }

            self.logger.error(str(exception), extra=details, exc_info=True)

            raise VTEXRequestError(**details) from None  # type: ignore[arg-type]

        self._raise_from_response(response=response, config=request_config)

        return cast(
            VTEXResponseType,
            (response_class or VTEXDataResponse).factory(response),
        )

    def _raise_from_response(self, response: Response, config: VTEXConfig) -> None:
        if response.is_error and config.raise_for_status_code:
            try:
                data = to_snake_case_deep(response.json(strict=False))
            except JSONDecodeError:
                data = response.text or HTTPStatus(response.status_code).phrase

            raise VTEXResponseError(
                data,
                method=str(response.request.method).upper(),
                url=str(response.request.url),
                request_headers=response.request.headers,
                status_code=response.status_code,
                data=data,
                response_headers=response.headers,
            ) from None



from typing import Any

from limits import parse

from .._dto import VTEXDataResponse, VTEXItemsResponse
from .base import BaseAPI


class PaymentsGatewayAPI(BaseAPI):
    """
    Client for the Payments Gateway API.
    https://developers.vtex.com/docs/api-reference/payments-gateway-api
    """

    ENVIRONMENT = "vtexpayments"

    def get_transaction(
        self,
        transaction_id: str,
        **kwargs: Any,
    ) -> VTEXDataResponse[Any]:
        return self._request(
            method="GET",
            endpoint=f"/api/pvt/transactions/{transaction_id}",
            environment=self.ENVIRONMENT,
            config=self.client.config.with_overrides(**kwargs),
            response_class=VTEXDataResponse[Any],
        )

    def list_transaction_interactions(
        self,
        transaction_id: str,
        **kwargs: Any,
    ) -> VTEXItemsResponse[Any, Any]:
        return self._request(
            method="GET",
            endpoint=f"/api/pvt/transactions/{transaction_id}/interactions",
            environment=self.ENVIRONMENT,
            config=self.client.config.with_overrides(**kwargs),
            rate_limit=parse("3500 per minute"),
            response_class=VTEXItemsResponse[Any, Any],
        )

    def list_transaction_payments(
        self,
        transaction_id: str,
        **kwargs: Any,
    ) -> VTEXItemsResponse[Any, Any]:
        return self._request(
            method="GET",
            endpoint=f"/api/pvt/transactions/{transaction_id}/payments",
            environment=self.ENVIRONMENT,
            config=self.client.config.with_overrides(**kwargs),
            response_class=VTEXItemsResponse[Any, Any],
        )

    def get_transaction_payment(
        self,
        transaction_id: str,
        payment_id: str,
        **kwargs: Any,
    ) -> VTEXDataResponse[Any]:
        return self._request(
            method="GET",
            endpoint=f"/api/pvt/transactions/{transaction_id}/payments/{payment_id}",
            environment=self.ENVIRONMENT,
            config=self.client.config.with_overrides(**kwargs),
            response_class=VTEXDataResponse[Any],
        )

    def get_transaction_capabilities(
        self,
        transaction_id: str,
        **kwargs: Any,
    ) -> VTEXDataResponse[Any]:
        return self._request(
            method="GET",
            endpoint=f"/api/pvt/transactions/{transaction_id}/capabilities",
            environment=self.ENVIRONMENT,
            config=self.client.config.with_overrides(**kwargs),
            response_class=VTEXDataResponse[Any],
        )

    def get_transaction_cancellations(
        self,
        transaction_id: str,
        **kwargs: Any,
    ) -> VTEXDataResponse[Any]:
        return self._request(
            method="GET",
            endpoint=f"/api/pvt/transactions/{transaction_id}/cancellations",
            environment=self.ENVIRONMENT,
            config=self.client.config.with_overrides(**kwargs),
            response_class=VTEXDataResponse[Any],
        )

    def get_transaction_refunds(
        self,
        transaction_id: str,
        **kwargs: Any,
    ) -> VTEXDataResponse[Any]:
        return self._request(
            method="GET",
            endpoint=f"/api/pvt/transactions/{transaction_id}/refunds",
            environment=self.ENVIRONMENT,
            config=self.client.config.with_overrides(**kwargs),
            response_class=VTEXDataResponse[Any],
        )

    def get_transaction_settlements(
        self,
        transaction_id: str,
        **kwargs: Any,
    ) -> VTEXDataResponse[Any]:
        return self._request(
            method="GET",
            endpoint=f"/api/pvt/transactions/{transaction_id}/settlements",
            environment=self.ENVIRONMENT,
            config=self.client.config.with_overrides(**kwargs),
            response_class=VTEXDataResponse[Any],
        )



from contextlib import contextmanager
from logging import CRITICAL, Logger, getLogger
from typing import Any, Callable, Generator, List, Union

from httpx import HTTPStatusError
from tenacity import RetryCallState


def get_logger(name: str, parent: Union[Logger, None] = None) -> Logger:
    if parent:
        return parent.getChild(name)

    return getLogger(f"vtex.{name}")


@contextmanager
def disable_loggers(logger_names: List[str]) -> Generator[None, None, None]:
    current_log_levels = {}

    for logger_name in logger_names:
        logger = getLogger(logger_name)
        current_log_levels[logger_name] = logger.level

        logger.setLevel(CRITICAL)

    try:
        yield
    finally:
        for logger_name, log_level in current_log_levels.items():
            getLogger(logger_name).setLevel(log_level)


def log_before_retry(
    logger: Logger,
    log_level: int,
    **kwargs: Any,
) -> Callable[[RetryCallState], None]:
    def retry_log(retry_state: RetryCallState) -> None:
        if not retry_state.outcome or not retry_state.next_action:
            raise RuntimeError("Retry log called before request was finished")

        exception = retry_state.outcome.exception()
        if not isinstance(exception, HTTPStatusError):
            raise RuntimeError("Retry log called without an http status error outcome")

        method = str(exception.request.method)
        url = str(exception.request.url)
        status_code = str(exception.response.status_code)
        reason = str(exception.response.reason_phrase)

        logger.log(
            log_level,
            f"Retrying {method} {url} in {retry_state.next_action.sleep}s as "
            f"attempt {retry_state.attempt_number} failed with: {status_code} {reason}",
            extra={
                "exception": exception,
                "attempt": retry_state.attempt_number,
                "sleep": retry_state.next_action.sleep,
                "method": method,
                "url": url,
                "status_code": status_code,
                **kwargs,
            },
        )

    return retry_log


CLIENT_LOGGER = get_logger(name="client")
EXCEPTIONS_LOGGER = get_logger(name="exceptions")
UTILS_LOGGER = get_logger(name="utils")



from logging import CRITICAL, DEBUG, ERROR, INFO, WARNING
from sys import version_info
from typing import Literal

if version_info >= (3, 12):
    from typing import TypedDict
else:
    from typing_extensions import TypedDict

HTTPMethodType = Literal[
    "DELETE",
    "GET",
    "HEAD",
    "OPTIONS",
    "PATCH",
    "POST",
    "PUT",
    "delete",
    "get",
    "head",
    "options",
    "patch",
    "post",
    "put",
]

LogLevelType = Literal[  # type: ignore[valid-type]
    DEBUG,
    INFO,
    WARNING,
    ERROR,
    CRITICAL,
]

OrderingDirectionType = Literal["ASC", "DESC", "asc", "desc"]

TypedDict = TypedDict



from typing import Any

from .._dto import VTEXDataResponse
from .base import BaseAPI


class PricingAPI(BaseAPI):
    """
    Client for the Pricing API.
    https://developers.vtex.com/docs/api-reference/pricing-api
    """

    def get_price(self, sku_id: str, **kwargs: Any) -> VTEXDataResponse[Any]:
        return self._request(
            method="GET",
            endpoint=f"/pricing/prices/{sku_id}",
            config=self.client.config.with_overrides(**kwargs),
            response_class=VTEXDataResponse[Any],
        )



from datetime import datetime
from typing import Any, List, Type, Union

from httpx._types import (
    CookieTypes,
    HeaderTypes,
    QueryParamTypes,
    RequestContent,
    RequestData,
    RequestFiles,
)

from .. import VTEXError
from .._dto import VTEXResponseType
from .._types import HTTPMethodType
from .._utils import to_datetime
from .base import BaseAPI
from .types.catalog import Seller
from .types.license_manager import AccountSite


class CustomAPI(BaseAPI):
    """
    Client for calling endpoints that have not yet been implemented by the SDK.
    You can directly call the `request` method to call any VTEX API.
    """

    def request(
        self,
        method: HTTPMethodType,
        endpoint: str,
        environment: Union[str, None] = None,
        headers: Union[HeaderTypes, None] = None,
        cookies: Union[CookieTypes, None] = None,
        params: Union[QueryParamTypes, None] = None,
        json: Union[Any, None] = None,
        data: Union[RequestData, None] = None,
        content: Union[RequestContent, None] = None,
        files: Union[RequestFiles, None] = None,
        response_class: Union[Type[VTEXResponseType], None] = None,
        **kwargs: Any,
    ) -> VTEXResponseType:
        return self._request(
            method=method,
            endpoint=endpoint,
            environment=environment,
            headers=headers,
            cookies=cookies,
            params=params,
            json=json,
            data=data,
            content=content,
            files=files,
            config=self.client.config.with_overrides(**kwargs),
            response_class=response_class,
        )

    def get_account_name(self) -> str:
        return self.client.license_manager.get_account().data["account_name"]

    def get_creation_date(self) -> datetime:
        return to_datetime(
            self.client.license_manager.get_account().data["creation_date"],
        ).replace(hour=0, minute=0, second=0, microsecond=0)

    def list_sites(self) -> List[AccountSite]:
        return self.client.license_manager.get_account().data["sites"]

    def get_main_seller(self) -> Seller:
        for seller in self.client.catalog.list_sellers(seller_type=1).items:
            if seller["seller_id"] == "1":
                return seller

        raise VTEXError("Could not find main seller")

    def list_market_place_sellers(self, include_inactive: bool = False) -> List[Seller]:
        return [
            seller
            for seller in self.client.catalog.list_sellers(seller_type=1).items
            if (include_inactive or seller["is_active"]) and seller["seller_id"] != "1"
        ]

    def list_franchise_sellers(self, include_inactive: bool = False) -> List[Seller]:
        return [
            seller
            for seller in self.client.catalog.list_sellers(seller_type=2).items
            if include_inactive or seller["is_active"]
        ]



from typing import List, Union

from ..._types import TypedDict
from .generic import PagePagination


class AccountAppKey(TypedDict):
    app_key: str
    id: str
    is_active: bool
    is_blocked: bool
    label: str


class AccountSite(TypedDict):
    name: str


class GetAccountData(TypedDict):
    account_name: str
    app_keys: List[AccountAppKey]
    company_name: str
    creation_date: str
    have_parent_account: bool
    id: str
    inactivation_date: Union[str, None]
    is_active: bool
    is_operating: bool
    name: str
    operation_date: Union[str, None]
    parent_account_id: Union[str, None]
    parent_account_name: Union[str, None]
    sites: List[AccountSite]
    trading_name: str


class UserRole(TypedDict):
    id: int
    name: str


class RoleProduct(TypedDict):
    name: str


class Role(TypedDict):
    id: int
    is_admin: bool
    name: str
    products: List[RoleProduct]
    role_type: int


class ListRolesData(TypedDict):
    items: List[Role]
    paging: PagePagination
